unit unitSyntaxCheck;

interface

uses unitSyntaxInformation, unitTokeniser, SysUtils, unitErrorHandler;

type
  // cetNone is used to mark the end of the TCAOSErrorData array
  TCAOSErrorType = (cetNone, cetTokenError, cetTypeMismatch, cetNotEnoughParams, cetTooManyParams, cetFirstTokenNotCommand);

  TCAOSErrorInfo = record
    description: string;
    tokenno: Integer; expected, actual: TCAOSType; // only relevant for cetTypeMismatch
    errortype: TCAOSErrorType;
  end;

  TCAOSErrorData = array[1..50] of TCAOSErrorInfo;

  ENotEnoughParams = class(Exception);

function SyntaxCheckTokenSet(t: TTokenSet): TCAOSErrorData;
function SyntaxCheckLine(s: string): TCAOSErrorData;

implementation

function isconditionjoin(s: string): boolean;
begin
  Result := (s = 'and') or (s = 'or');
end;

// where TOKEN is one of the following: eq ne gt ge lt le = <> > >= < <=
function isconditiontoken(s: string): boolean;
begin
  Result := False;
  if s = 'eq' then Result := True
  else if s = 'ne' then Result := True
  else if s = 'gt' then Result := True
  else if s = 'ge' then Result := True
  else if s = 'lt' then Result := True
  else if s = 'le' then Result := True
  else if s = '=' then Result := True
  else if s = '<>' then Result := True
  else if s = '>' then Result := True
  else if s = '>=' then Result := True
  else if s = '<' then Result := True
  else if s = '<=' then Result := True;
end;

function SyntaxCheckTokenSet(t: TTokenSet): TCAOSErrorData;
var
  i, noerrors: Integer;
  firstkey: Boolean;

procedure ParseKeyword(j: TTokenInfo); forward;

procedure StripConditionParam;
begin
  if i > t.notokens then begin
    Inc(noerrors);
    Result[noerrors].errortype := cetNotEnoughParams;
    i := t.notokens;
    raise Exception.Create('this is caught by caller (syntaxchecktokenset->stripconditionparam)');
  end;
  if (t.tokens[i].tokentype = ctUnknown) or (t.tokens[i].tokentype = ctCommand) then begin
    Inc(noerrors);
    Result[noerrors].errortype := cetTypeMismatch;
    Result[noerrors].tokenno := i;
    Result[noerrors].actual := t.tokens[i].tokentype;
    Result[noerrors].expected := ctAnything;
  end;
  if t.tokens[i].iskeyword then begin
    ParseKeyword(t.tokens[i]);
  end;
end;

procedure StripConditionSegment;
begin
  // syntax of a condition: unknown TOKEN unknown [AND|OR unknown TOKEN unknown]
  // where TOKEN is one of the following: eq ne gt ge lt le = <> > >= < <=

  try StripConditionParam; except on Exception do Exit; end;
  Inc(i);
  if i > t.notokens then begin
    Inc(noerrors);
    Result[noerrors].errortype := cetNotEnoughParams;
    i := t.notokens;
    Exit;
  end;
  if not isconditiontoken(t.tokens[i].stringvalue) then begin
    Inc(noerrors);
    Result[noerrors].errortype := cetTypeMismatch;
    Result[noerrors].tokenno := i;
    Result[noerrors].actual := t.tokens[i].tokentype;
    Result[noerrors].expected := ctCondition;
    Exit;
  end;
  Inc(i);
  try StripConditionParam; except on Exception do Exit; end;

  if i < t.notokens then if isconditionjoin(t.tokens[i + 1].stringvalue) then begin
    i := i + 2;
    StripConditionSegment;
  end;
end;

procedure ParseKeyword(j: TTokenInfo);
var
  k: Integer;
begin
  if firstkey and (j.tokentype <> ctCommand) then begin
    Inc(noerrors);
    Result[noerrors].errortype := cetFirstTokenNotCommand;
    Exit;
  end;

  firstkey := false;
  if not j.iskeyword then begin
    if not firstkey then ErrorDialog(18, True)
    else ErrorDialog(19, True);
    Exit;
  end;

  k := 1;
  while k <= j.keywordinfo.noparams do begin
    Inc(i);
    if i > t.notokens then begin
      Inc(noerrors);
      Result[noerrors].errortype := cetNotEnoughParams;
      raise ENotEnoughParams.Create('not enough params');
    end;

    if j.keywordinfo.paramtypes[k] = ctCondition then
      StripConditionSegment
    else if not tokenmatch(t.tokens[i].tokentype, j.keywordinfo.paramtypes[k]) then begin
      Inc(noerrors);
      Result[noerrors].errortype := cetTypeMismatch;
      Result[noerrors].tokenno := i;
      Result[noerrors].actual := t.tokens[i].tokentype;
      Result[noerrors].expected := j.keywordinfo.paramtypes[k];
    end;
    if (t.tokens[i].iskeyword) and (j.keywordinfo.paramtypes[k] <> ctLabel) then begin // labels often have a name like 'face', and they're detected as a keyword
      ParseKeyword(t.tokens[i]);
    end;
    Inc(k);
  end;
end;

begin
  noerrors := 0; i := 1; firstkey := true; // (start at token 2)

  if t.notokens = 0 then begin Result[1].errortype := cetNone; Exit; end; // empty line

  try
    ParseKeyword(t.tokens[1]);
  except
    on ENotEnoughParams do;
  end;

  if i < t.notokens then begin
    Inc(noerrors);
    Result[noerrors].errortype := cetTooManyParams;
  end;

  Result[noerrors + 1].errortype := cetNone;
end;

function SyntaxCheckLine(s: string): TCAOSErrorData;
var
  t: TTokenSet;
begin
  t := TTokenSet.Create;
  t.SlurpTokenSet(s);
  Result := SyntaxCheckTokenSet(t);
  t.Free;
end;

end.
