unit unitSyntaxInformation;

// CAOS Syntax Routines

interface

uses SysUtils, Classes;

type
  // ctUnknown means its not a keyword, integer, float, string or byte-string
  // ctAnything is only used in condition errors ... :P
  // ctLabel is a label, but the code which handles it is a kludge right now
  TCAOSType = (ctUnknown, ctInteger, ctFloat, ctString, ctByteString, ctAgent, ctVariable, ctCommand, ctCondition, ctAnything, ctNotCommand, ctLabel);

  TKeywordInfo = class
  public
    name, shortdesc, category, longdesc: string;
    ktype: TCAOSType;
    noparams: Integer;
    paramnames: array[1..30] of string;
    paramtypes: array[1..30] of TCAOSType;

    blockstart, blockmiddle, blockend, subroutine: Boolean;
    scriptend, installscriptbegin, removescriptbegin, scriptbegin: Boolean;

    procedure ParseFrom(s: String);
  end;

  TSyntaxData = class
  public
    // the variables really should be private ... but :(
    keywords: array[1..5000] of TKeywordInfo;
    count: integer;
    multiplekeywords: TStringList;
    
    procedure ReadFromFile(filename: string);
    function GetKeywordData(s: string; neededtype: TCAOSType): TKeywordInfo;
    function IsMultipleKeyword(s: string): boolean;
  end;

var
  syntaxdata: TSyntaxData;

function typetostring(a: TCAOSType): string;
function tokenmatch(actual, expected: TCAOSType): Boolean;
function GrabType(s: string): TCAOSType;

implementation

function typetostring(a: TCAOSType): string;
begin
  if a = ctInteger then Result := 'integer'
  else if a = ctFloat then Result := 'float'
  else if a = ctCommand then Result := 'command'
  else if a = ctAgent then Result := 'agent'
  else if a = ctString then Result := 'string'
  else if a = ctByteString then Result := 'byte-string'
  else if a = ctUnknown then Result := 'unknown'
  else if a = ctVariable then Result := 'variable'
  else if a = ctCondition then Result := 'condition'
  else if a = ctAnything then Result := 'anything'
  else if a = ctLabel then Result := 'label'
  else Result := '**error**';
end;

function tokenmatch(actual, expected: TCAOSType): Boolean;
begin
  Result := False;
  if actual = expected then Result := True
  else if actual = ctCommand then Result := False
  else if (expected = ctLabel) then Result := True // fixme: this should be fixed up in syntax check code (this code here will let 'gsub 5' through!)
  else if expected = ctNotCommand then Result := True
  else if (actual = ctInteger) and (expected = ctFloat) then Result := True // allow an integer to be a float
  else if (actual = ctFloat) and (expected = ctInteger) then Result := True // this may be a kludge :(
  else if actual = ctVariable then Result := True // a variable can contain anything
  else if (expected = ctAnything) and (actual <> ctUnknown) then Result := True
  else if expected = ctUnknown then Result := True; // if the expected type is unknown, allow anything
end;

// RemoveSpecial assists in splitting a string up which is in char-seperated segments, eg comma-delimited
function RemoveSpecial(var fromm: string; c: Char): String;
begin
  if pos(c, fromm) = 0 then
  begin
    Result := fromm;
    fromm := '';
    Exit;
  end;
  Result := copy(fromm, 1, pos(c, fromm) - 1);
  fromm := copy(fromm, pos(c, fromm) + 1, length(fromm) - pos(c, fromm));
end;

function GrabType(s: string): TCAOSType;
begin
  if s = 'variable' then Result := ctVariable
  else if s = 'condition' then Result := ctCondition
  else if s = 'integer' then Result := ctInteger
  else if s = 'float' then Result := ctFloat
  else if s = 'decimal' then Result := ctFloat // fixme: should this be true?
  else if s = 'string' then Result := ctString
  else if s = 'byte-string' then Result := ctByteString
  else if s = 'command' then Result := ctCommand
  else if s = 'agent' then Result := ctAgent
  else if s = 'anything' then Result := ctAnything
  else if s = 'label' then Result := ctLabel
  else Result := ctUnknown;
end;

// *** TKeywordInfo ***

procedure TKeywordInfo.ParseFrom(s: String);
var
  i: integer;
begin
  name := LowerCase(RemoveSpecial(s, ','));
  shortdesc := RemoveSpecial(s, ',');
  ktype := GrabType(RemoveSpecial(s, ','));
  noparams := StrToInt(RemoveSpecial(s, ','));
  for i := 1 to noparams do begin
    paramtypes[i] := GrabType(RemoveSpecial(s, ','));
    paramnames[i] := RemoveSpecial(s, ',');
  end;
  i := StrToInt(RemoveSpecial(s, ','));
  blockstart := (i and 1) = 1;
  blockmiddle := (i and 2) = 2;
  blockend := (i and 4) = 4;
  subroutine := (i and 8) = 8;
  scriptend := (i and 16) = 16;
  installscriptbegin := (i and 32) = 32;
  removescriptbegin := (i and 64) = 64;
  scriptbegin := (i and 128) = 128;
  category := RemoveSpecial(s, ',');
  longdesc := RemoveSpecial(s, ',');
end;

// *** TSyntaxData ***

procedure TSyntaxData.ReadFromFile(filename: string);

procedure ParseLine(p: string);
var
  s: string;
begin
  if Length(p) = 0 then Exit;
  if p[1] = '/' then Exit;
  if p[1] = '#' then begin
    s := Copy(p, 2, Length(p) - 1);
    multiplekeywords.Add(s);
    Exit;
  end;

  Inc(count);
  keywords[count] := TKeywordInfo.Create;
  keywords[count].ParseFrom(p);
end;

var
  st: TStringList;
  i: Integer;
begin
  st := TStringList.Create;
  multiplekeywords := TStringList.Create;
  try
    st.LoadFromFile(filename);
    for i := 0 to (st.Count - 1) do
      ParseLine(st[i]);
  finally
    st.Free;
  end;
end;

function TSyntaxData.GetKeywordData(s: string; neededtype: TCAOSType): TKeywordInfo;
var
  i: Integer;
begin
  Result := nil;
  s := LowerCase(s);
  for i := 1 to Count do begin
    if keywords[i].name = s then if tokenmatch(keywords[i].ktype, neededtype) then Result := keywords[i];
  end;
end;

function TSyntaxData.IsMultipleKeyword(s: string): boolean;
begin
  Result := multiplekeywords.IndexOf(LowerCase(s)) <> -1;
end;

initialization
  syntaxdata := TSyntaxData.Create;
end.