unit unitTokeniser;

// CAOS Tokenising Routines

interface

uses SysUtils, Classes, unitSyntaxInformation;

type
  // exceptions thrown by SlurpToken
  ECAOSOpenString = class(Exception); // reached end of line but string is still open (missing quotemark/bracket)
  ECAOSStringClosed = class(Exception); // got a string end but it isn't followed by an end, space or tab
  ECAOSInvalidChar = class(Exception); // invalid character at this point
  ECAOSEmptyString = class(Exception); // this means the string passed to SlurpToken was empty, *not* a CAOS string :)
  ECAOSInvalidEscape = class(Exception); // invalid escape, or escape not closed, or escape outside string

  // TTokenInfo represents a single token. You can either fill it in yourself (you're mad) or use SlurpToken
  // to slurp a token out of a string.
  TTokenInfo = class
  public
    tokentype: TCAOSType; // the type of the token
    iskeyword: Boolean; // is this token a keyword?
    keywordinfo: TKeywordInfo; // this is only valid for keywords
    intvalue: Integer; // this is only valid for integers
    floatvalue: Double; // this is valid for floats and integers
    stringvalue: String; // this is always valid

    // this isn't brilliant, but its better than my convolted NextToken =)
    // it does NOT have special handling for comments, nor check whether byte strings are valid or not
    // it throws exceptions for some errors, you'll just have to skip the rest of the line :)
    function SlurpToken(s: string): string;
  end;

  // TTokenSet represents a set of tokens. Use SlurpTokenSet to slurp it out of a string, and you can then
  // pass the TTokenSet to whatever you want ... :)
  TTokenSet = class
    tokens: array[1..150] of TTokenInfo; // corrected 23/04/00 by Fuzzie, array was too small at 50
    notokens: Integer;

    errorinfo: string; // error information for this line

    procedure SlurpTokenSet(s: string);
  end;

var
//  internal_cmd{, internal_usecmd}: boolean;
  internal_cmd: TCAOSType;

implementation

procedure TTokenSet.SlurpTokenSet(s: string);
begin
  notokens := 0;
  errorinfo := '';
  while s <> '' do begin
    // fixme: we need to find out the expected type here
    if notokens = 0 then internal_cmd := ctCommand
    // following line is kludge for 'setv xxxx face', which shows up a lot
    // the problem is that the tokeniser picks up the string variant of face =(
    else if (LowerCase(Trim(s)) = 'face') and (LowerCase(tokens[1].stringvalue) = 'setv') then internal_cmd := ctInteger
    else internal_cmd := ctNotCommand;

    Inc(notokens);
    tokens[notokens] := TTokenInfo.Create;
//    try
      s := tokens[notokens].SlurpToken(s);
//    except
      // Fuzzie removed this 02/04/00 - colorcoding and syntaxcheck catch exceptions now
      // if you're intending to use the test form, you'll need to restore it
      // fixme: fix the test form
      // on e: Exception do begin errorinfo := e.Message; Exit; end;
//    end;
  end;

  internal_cmd := ctAnything; // reset it
end;

// FIXME: this function needs OPTIMISING desperately! and I mean desperately
// color-coding depends on it, and I'm sure you can see how slow that is :(
function TTokenInfo.SlurpToken(s: string): string;
var
  p: string;
  i: Integer;
  c: Char;
  stringopen, bytestringopen: boolean;
  escapeopen: boolean;

procedure DoIt(z: string);
var
  k: TKeywordInfo;
  p: TTokenInfo;
begin
  // work out what the token is

  iskeyword := false;

  stringvalue := z;
  try
    intvalue := StrToInt(z); tokentype := ctInteger; Exit;
  except
    on EConvertError do begin
      // not an integer
      try
        floatvalue := StrToFloat(z); tokentype := ctFloat; Exit;
      except
        on EConvertError do begin
          // not an integer or a float - so it must be a string, or a command
          // we know at this point that we don't have a zero-length string, so its safe to access [1]
          if s[1] = '"' then begin
            tokentype := ctString; Exit;
          end else if s[1] = '[' then begin
            tokentype := ctByteString; Exit;
          end else begin
            // it isn't a string, a byte-string, an integer or a float
            // so it must be either a keyword or wrong =D

            if syntaxdata.ismultiplekeyword(LowerCase(z)) then begin
              // we have a multiple keyword here, folks
              // z contains the first portion ... we need to use another SlurpToken to slurp the second token :)
              p := TTokenInfo.Create;
              p.SlurpToken(Copy(s, i + 1, Length(s) - (i - 1)));
              z := z + ' ' + p.stringvalue;
              stringvalue := z; // reset :)
              p.Free;
              i := i + 5; // skip 5 chars
            end;

            k := syntaxdata.GetKeywordData(LowerCase(z), internal_cmd);

            if k <> nil then begin
              iskeyword := true;
              TokenType := k.ktype;
              keywordinfo := k;
              Exit;
            end;
          end;
        end;
      end;
    end;
  end;

  tokentype := ctUnknown;
end;

begin
  p := ''; stringopen := false; bytestringopen := false; escapeopen := false;
  s := Trim(s); // trim it of leading/trailing spaces/tabs
  for i := 1 to Length(s) do begin
    // set c to the current letter
    c := s[i];
    if escapeopen then begin
      escapeopen := false;
      if c <> 'n' then // fixme: what do we do about \n escapes?
        if (c <> '"') and (c <> '\') then raise ECAOSInvalidEscape.Create('invalid escape ' + c);
    end else if c = '\' then begin
      if not stringopen then raise ECAOSInvalidEscape.Create('escape not allowed outside string');
      escapeopen := true;
    end else if (not stringopen) and (not bytestringopen) then begin
      // if we're not in a string, check for tabs and spaces
      if c = ' ' then begin DoIt(p); Break; end; // space
      if c = #9 then begin DoIt(p); Break; end; // tab
      if c = '"' then begin // else, see if we need to start a string
        if i <> 1 then raise ECAOSInvalidChar.Create('unexpected "');
        stringopen := true; end;
      if c = '[' then begin // else, see if we need to start a string
        if i <> 1 then raise ECAOSInvalidChar.Create('unexpected [');
        bytestringopen := true; end;
      if c = ']' then raise ECAOSInvalidChar.Create('unexpected ]');
    end else if stringopen then begin
      // if we're in a string, check for an end-string
      if c = '"' then begin
        if (Length(s) > i) and ((s[i + 1] <> ' ') and (s[i + 1] <> #9)) then
          raise ECAOSStringClosed.Create('string closed but token continues');
        stringopen := false;
        p := p + c;
        DoIt(p);
        Break;
      end;
    end else if bytestringopen then begin
      // if we're in a string, check for an end-string
      if c = ']' then begin if (Length(s) > i) and ((s[i + 1] <> ' ') and (s[i + 1] <> #9)) then raise ECAOSStringClosed.Create('string closed but token continues'); bytestringopen := false; p := p + c; DoIt(p); Break; end;
    end;
    p := p + c;
    if i = Length(s) then
      DoIt(p);
  end;
  if p = '' then raise ECAOSEmptyString.Create('empty string');
  if stringopen then raise ECAOSOpenString.Create('open string');
  if bytestringopen then raise ECAOSOpenString.Create('open bytestring');
  if escapeopen then raise ECAOSInvalidEscape.Create('open escape');
  Result := Copy(s, i + 1, Length(s) - (i - 1));
end;

initialization
  internal_cmd := ctAnything;
//  internal_usecmd := false;
end.
